import contextlib
import glob
import os
import re
from subprocess import Popen, PIPE, STDOUT

import psutil

from . import configuration
from .configuration import PlotmanConfig


def is_plot_moving(cmdline: list) -> bool:
    return (
            len(cmdline) > 3
            and cmdline[0].endswith('plmo')
    )


def is_plmo_nfs(cmdline: list) -> bool:
    test = " ".join(cmdline)
    m = re.match(r"^\/mnt\/nfs.*.?(\d+).\/", test)
    if m:
        return True
    else:
        return False


class FarmPlot:
    """
    The final state to move files from tmp folder to given network hard drive devices
    """

    def __init__(self, cfg: PlotmanConfig):
        self.dsts = cfg.directories.dst
        self.checktmps = cfg.directories.tmp
        self.schedule = cfg.scheduling.polling_time_s
        self.log_file_path = configuration.get_log_path()

    @staticmethod
    def get_running_moving_jobs() -> list:
        jobs = []

        for proc in psutil.process_iter(['pid', 'cmdline']):
            # Ignore processes which most likely have terminated between the time of iteration and data access.
            with contextlib.suppress(psutil.NoSuchProcess, psutil.AccessDenied):
                if is_plot_moving(proc.cmdline()):
                    job = " ".join(proc.cmdline())
                    jobs.append(job)

        return jobs

    @staticmethod
    def get_nfs_details() -> list:
        nfs_list = []

        for proc in psutil.process_iter(['pid', 'cmdline']):
            # Ignore processes which most likely have terminated between the time of iteration and data access.
            with contextlib.suppress(psutil.NoSuchProcess, psutil.AccessDenied):
                if is_plot_moving(proc.cmdline()):
                    if is_plmo_nfs(proc.cmdline()):
                        test = " ".join(proc.cmdline())
                        g = re.match('(\d+)', test)
                        nfs_list.append(g[0])

        return nfs_list

    def start_nfs(self) -> None:

        count1 = len(glob.glob1("~/chia-blockchain", "plmo"))
        if count1 >= 1:
            for d in self.dsts:
                p = Popen(["./plmo", self.checktmps, d, 50000000, self.log_file_path], stdout=PIPE, stderr=STDOUT)
                output, error = p.communicate()
                output = output.strip().decode("utf-8")
                print("start moving file from temp folder to network FS farm location {}".format(d))
                if p.returncode:
                    print(f"E: {output}")
        else:
            print("Did not find the plmo executable.")

    def start_nvme_local_dir(self) -> None:
        subs = [x[0] for x in os.walk("/mnt")]
        count2 = len(subs)
        for rsub in subs:
            print(rsub)



# TODO: migrate away from unittest patch
from unittest.mock import patch

from plotmanx import reporting


def test_phases_str_basic():
    assert (reporting.phases_str([(1, 2), (2, 3), (3, 4), (4, 0)]) ==
            '1:2 2:3 3:4 4:0')


def test_phases_str_elipsis_1():
    assert (reporting.phases_str([(1, 2), (2, 3), (3, 4), (4, 0)], 3) ==
            '1:2 [+1] 3:4 4:0')


def test_phases_str_elipsis_2():
    assert (reporting.phases_str([(1, 2), (2, 3), (3, 4), (4, 0)], 2) ==
            '1:2 [+2] 4:0')


def test_phases_str_none():
    assert (reporting.phases_str([(None, None), (2, None), (3, 0)]) ==
            '?:? 2:? 3:0')


def test_job_viz_empty():
    assert (reporting.job_viz([]) == '1:        |2...3:       |4: ')


@patch('plotmanx.job.Job')
def job_w_phase(ph, MockJob):
    j = MockJob()
    j.progress.return_value = ph
    return j


def test_job_viz_positions():
    jobs = [job_w_phase((1, 1)),
            job_w_phase((2, 0)),
            job_w_phase((2, 4)),
            job_w_phase((2, 7)),
            job_w_phase((4, 0))]

    assert (reporting.job_viz(jobs) == '1: 1      |2...3:       |4:1')


def test_job_viz_counts():
    jobs = [job_w_phase((2, 2)),
            job_w_phase((2, 3)),
            job_w_phase((2, 3)),
            job_w_phase((2, 4)),
            job_w_phase((2, 4)),
            job_w_phase((2, 4)),
            job_w_phase((2, 5)),
            job_w_phase((2, 5)),
            job_w_phase((2, 5)),
            job_w_phase((2, 5)),
            job_w_phase((3, 1)),
            job_w_phase((3, 1)),
            job_w_phase((3, 1)),
            job_w_phase((3, 1)),
            job_w_phase((3, 1)),
            job_w_phase((3, 1)),
            ]

    assert (reporting.job_viz(jobs) == '1:        |2...3: 5     |4: ')



#!/usr/bin/python
#

from datetime import datetime

import pkg_resources
import requests
from tornado import web, ioloop, httpserver

from .sql import SQLX
from ..configuration import PlotmanConfig, get_dash_v1, get_dash_v1_static

try:
    import simplejson as json
except ImportError:
    import json


def PostDat(dp: dict, cfg: PlotmanConfig):
    # sending post request and saving response as response object
    payload = json.dumps(dp)
    requests.post(url=f'http://{cfg.apis.target}:{cfg.apis.port}/report', data=payload)


class ApiBase(web.RequestHandler):
    SUPPORTED_METHODS = ["GET"]

    def initialize(self, *args, **kwargs):
        self.remote_ip = self.request.headers.get('X-Forwarded-For', self.request.headers.get('X-Real-Ip', self.request.remote_ip))
        self.using_ssl = (self.request.headers.get('X-Scheme', 'http') == 'https')

    def set_default_headers(self):
        """Set the default response header to be JSON."""
        self.set_header("Content-Type", 'application/json; charset="utf-8"')

    def send_response(self, data, status=200):
        """Construct and send a JSON response with appropriate status code."""
        self.set_status(status)
        t = int(datetime.now().timestamp()),
        response = {'result': 1, 'data': data, 'ts': t}
        self.write(json.dumps(response))


"""

from prometheus_client.core import GaugeMetricFamily

class PlotmanCollector:
    def collect(self):
        cfg = configuration.get_validated_configs()
        jobs = Job.genTasks(cfg.directories.log)
        count = len(sorted(jobs, key=job.Job.get_time_wall))
        yield GaugeMetricFamily("plotman_jobs_count", "Number of plotting jobs running", value=count)

if __name__ == "__main__":
    start_http_server(8001)
    REGISTRY.register(PlotmanCollector())
    while True:
        time.sleep(1)
"""


class NodeHandle(web.RequestHandler):
    def initialize(self, *args, **kwargs):
        self.remote_ip = self.request.headers.get('X-Forwarded-For', self.request.headers.get('X-Real-Ip', self.request.remote_ip))
        self.using_ssl = (self.request.headers.get('X-Scheme', 'http') == 'https')

    def post(self):
        self.set_header("Content-Type", "text/plain")
        req_body = self.request.body
        sq = SQLX()
        sq.schema_plan()

        jpayload = dict()

        try:
            jpayload = json.loads(req_body)
        except:
            print("error from decoding json file.")
            return

        try:
            ip_remo = jpayload['identity']
        except:
            ip_remo = self.request.headers.get("X-Real-IP") or \
                      self.request.headers.get("X-Forwarded-For") or \
                      self.request.remote_ip

        tstime = datetime.now().strftime('%m-%d %H:%M:%S')

        if req_body is not None:
            sq.datainput(jpayload, ip_remo, tstime)

        sq.end()

        print("Report data ------")
        print(f"remote ip: {ip_remo} and ver. {jpayload['version']}")


class DashSimpleListNodes(ApiBase):

    def get(self):
        sq = SQLX()
        res = sq.getNodes()
        sq.end()
        # key = self.get_argument('key', None)
        self.send_response(res)


class ApiV1Review(ApiBase):

    def get(self):
        """List of routes for this API."""
        routes = {
            'info': 'GET /api/v1',
            'register': 'POST /api/v1/accounts',
            'single profile detail': 'GET /api/v1/accounts/<username>',
            'edit profile': 'PUT /api/v1/accounts/<username>',
            'delete profile': 'DELETE /api/v1/accounts/<username>',
            'login': 'POST /api/v1/accounts/login',
            'logout': 'GET /api/v1/accounts/logout',
            "user's tasks": 'GET /api/v1/accounts/<username>/tasks',
            "create task": 'POST /api/v1/accounts/<username>/tasks',
            "task detail": 'GET /api/v1/accounts/<username>/tasks/<id>',
            "task update": 'PUT /api/v1/accounts/<username>/tasks/<id>',
            "delete task": 'DELETE /api/v1/accounts/<username>/tasks/<id>'
        }
        self.write(json.dumps(routes))


class FrontEndLoad(web.RequestHandler):
    def initialize(self, *args, **kwargs):
        self.remote_ip = self.request.headers.get('X-Forwarded-For', self.request.headers.get('X-Real-Ip', self.request.remote_ip))
        self.using_ssl = (self.request.headers.get('X-Scheme', 'http') == 'https')

    def get(self):
        ver = pkg_resources.get_distribution('plotmanx').version
        self.set_header("Content-Type", "text/plain")
        self.render("index.html", messages=None, version=ver)


def start_master_api_node(cfg: PlotmanConfig):
    try:
        version = pkg_resources.get_distribution('plotmanx').version
        print(f"Now plotman API - v{version}")
        print(f"API running port {cfg.apis.port} is now listening.. ")

        settings = dict(
            debug=True,
            template_path=get_dash_v1(),
            static_path=get_dash_v1_static()
        )

        appcli = web.Application([
            (r"/report", NodeHandle),
            (r"/nodes", DashSimpleListNodes),
            (r"/apiv1", ApiV1Review),
            (r"/", FrontEndLoad),
        ], **settings)

        http_server = httpserver.HTTPServer(appcli)
        http_server.listen(cfg.apis.port)
        ioloop.IOLoop.current().start()

    except:
        print("Port is in use. Please close the daemon app and start again. ")



import csv
import sys

from dateutil.parser import parse as parse_date
from .parser import PlotLogParser
from .plotinfo import PlotInfo


def export(logfilenames, save_to=None):
    if save_to is None:
        send_to_stdout(logfilenames)
    else:
        save_to_file(logfilenames, save_to)


def save_to_file(logfilenames, filename: str):
    with open(filename, 'w') as file:
        generate(logfilenames, file)


def send_to_stdout(logfilenames):
    generate(logfilenames, sys.stdout)


def header(writer):
    writer.writerow([
        'Plot ID',
        'Started at',
        'Date',
        'Size',
        'Buffer',
        'Buckets',
        'Threads',
        'Tmp dir 1',
        'Tmp dir 2',
        'Phase 1 duration (raw)',
        'Phase 1 duration',
        'Phase 1 duration (minutes)',
        'Phase 1 duration (hours)',
        'Phase 2 duration (raw)',
        'Phase 2 duration',
        'Phase 2 duration (minutes)',
        'Phase 2 duration (hours)',
        'Phase 3 duration (raw)',
        'Phase 3 duration',
        'Phase 3 duration (minutes)',
        'Phase 3 duration (hours)',
        'Phase 4 duration (raw)',
        'Phase 4 duration',
        'Phase 4 duration (minutes)',
        'Phase 4 duration (hours)',
        'Total time (raw)',
        'Total time',
        'Total time (minutes)',
        'Total time (hours)',
        'Copy time (raw)',
        'Copy time',
        'Copy time (minutes)',
        'Copy time (hours)',
        'Filename'
    ])


def parse_logs(logfilenames):
    parser = PlotLogParser()
    result = []

    for filename in logfilenames:
        info = parser.parse(filename)

        if not info.is_empty():
            result.append(info)

    result.sort(key=log_sort_key)
    return result


def log_sort_key(element: PlotInfo):
    return parse_date(element.started_at).replace(microsecond=0).isoformat()


def generate(logfilenames, out):
    writer = csv.writer(out)
    header(writer)
    logs = parse_logs(logfilenames)

    for info in logs:
        writer.writerow([
            info.plot_id,
            info.started_at,
            parse_date(info.started_at).strftime('%Y-%m-%d'),
            info.plot_size,
            info.buffer,
            info.buckets,
            info.threads,
            info.tmp_dir1,
            info.tmp_dir2,
            info.phase1_duration_raw,
            info.phase1_duration,
            info.phase1_duration_minutes,
            info.phase1_duration_hours,
            info.phase2_duration_raw,
            info.phase2_duration,
            info.phase2_duration_minutes,
            info.phase2_duration_hours,
            info.phase3_duration_raw,
            info.phase3_duration,
            info.phase3_duration_minutes,
            info.phase3_duration_hours,
            info.phase4_duration_raw,
            info.phase4_duration,
            info.phase4_duration_minutes,
            info.phase4_duration_hours,
            info.total_time_raw,
            info.total_time,
            info.total_time_minutes,
            info.total_time_hours,
            info.copy_time_raw,
            info.copy_time,
            info.copy_time_minutes,
            info.copy_time_hours,
            info.filename
        ])



class InvalidArgumentException(Exception):
    pass


class InvalidChiaLocationException(Exception):
    pass


class InvalidConfigurationSetting(Exception):
    pass


class InvalidYAMLConfigException(Exception):
    pass


class ManagerError(Exception):
    pass


class MissingImportError(Exception):
    pass


class TerminationException(Exception):
    pass


class TerminalTooSmallError(Exception):
    pass



import importlib.resources
import os

import pytest
from plotmanx import resources

from ..analyze.pri import LogFile


@pytest.fixture(name='logfile_path')
def logfile_fixture(tmp_path):
    log_name = '2021-05-07T20_25_52.851358+08_00.log'
    log_contents = importlib.resources.read_binary(resources, log_name)
    log_file_path = tmp_path.joinpath(log_name)
    log_file_path.write_bytes(log_contents)
    return log_file_path


def test_one():
    log_name = '2021-05-07T20_25_52.851358+08_00.log'
    dir = '/Users/hesdx/Documents/ipfs/plotman/src/plotmanx/_tests/resources'
    file = os.path.join(dir, log_name)
    fileo = LogFile(file)
    fileo.init_logfile()
    fileo.print()


test_one()



import math
import os
import re
import socket

import pendulum

GB = 1_000_000_000

KSIZE = {
    "32": 256.6,
    "33": 550,
    "34": 1118,
    "35": 2335,
}

PROGRESS_BAR = {
    "phase1_line_end": 801,
    "phase2_line_end": 834,
    "phase3_line_end": 2474,
    "phase4_line_end": 2620,
    "phase1_weight": 33.4,
    "phase2_weight": 20.43,
    "phase3_weight": 42.29,
    "phase4_weight": 3.88,
}


def is_phase_start(phase_pair) -> bool:
    (ph, subph) = phase_pair
    return ph < 3 and subph < 6


def phase_str(phase_pair) -> str:
    (ph, subph) = phase_pair
    return ((str(ph) if ph is not None else '?') + ':'
            + (str(subph) if subph is not None else '?'))


def abbr_path(path, putative_prefix) -> str:
    if putative_prefix and path.startswith(putative_prefix):
        return os.path.relpath(path, putative_prefix)
    else:
        return path


def df_b(d) -> int:
    'Return free space for directory (in bytes)'
    stat = os.statvfs(d)
    return stat.f_frsize * stat.f_bavail


def checkTempSpace(knm: int, tmpdir: str) -> bool:
    for k, v in KSIZE.items():
        if int(k) == knm:
            gb_free = df_b(tmpdir) / GB
            if gb_free > float(v):
                return True
    return False


def isSpaceCritical(tmpdir: str) -> bool:
    gb_free = df_b(tmpdir) / GB
    if gb_free < 0.1:
        return True
    else:
        return False


def availableSpace(tmpdir: str) -> str:
    gb_free = df_b(tmpdir) / GB
    return f"{gb_free} GB"


def get_k32_plotsize():
    return 108 * GB


def human_format(num, precision):
    magnitude = 0
    while abs(num) >= 1000:
        magnitude += 1
        num /= 1000.0
    return (('%.' + str(precision) + 'f%s') %
            (num, ['', 'K', 'M', 'G', 'T', 'P'][magnitude]))


def time_format(sec):
    if sec is None:
        return '-'
    if sec < 60:
        return '%ds' % sec
    else:
        return '%d:%02d' % (int(sec / 3600), int((sec % 3600) / 60))


def tmpdir_phases_str(tmpdir_phases_pair):
    tmpdir = tmpdir_phases_pair[0]
    phases = tmpdir_phases_pair[1]
    phase_str = ', '.join(['%d:%d' % ph_subph for ph_subph in sorted(phases)])
    return ('%s (%s)' % (tmpdir, phase_str))


def getMaxFitablePlots(drive_size: int, final_plot_size: int) -> float:
    driveSpaceBuffer = 0.05
    amt = ((drive_size - (drive_size * driveSpaceBuffer))) / final_plot_size
    return amt


def availableSizeDriveForPlots(sized: int, final_plot_size: int) -> float:
    return sized / final_plot_size


def split_path_prefix(items):
    if not items:
        return ('', [])

    prefix = os.path.commonpath(items)
    if prefix == '/':
        return ('', items)
    else:
        remainders = [os.path.relpath(i, prefix) for i in items]
        return (prefix, remainders)


def list_k32_plots(d):
    """List completed k32 plots in a directory (not recursive)"""
    plots = []
    for plot in os.listdir(d):
        if re.match(r'^plot-k32-.*plot$', plot):
            plot = os.path.join(d, plot)
            try:
                if os.stat(plot).st_size > (0.95 * get_k32_plotsize()):
                    plots.append(plot)
            except FileNotFoundError:
                continue

    return plots


def column_wrap(items, n_cols, filler=None):
    '''Take items, distribute among n_cols columns, and return a set
       of rows containing the slices of those columns.'''
    rows = []
    n_rows = math.ceil(len(items) / n_cols)
    for row in range(n_rows):
        row_items = items[row:: n_rows]
        # Pad and truncate
        rows.append((row_items + ([filler] * n_cols))[:n_cols])
    return rows


def is_freezed(job):
    return 'YES' if job.last_updated_time_in_min > 60 else 'NO'


def parse_chia_plot_time(s):
    # This will grow to try ISO8601 as well for when Chia logs that way
    return pendulum.from_format(s, 'ddd MMM DD HH:mm:ss YYYY', locale='en', tz=None)


def get_ip_address(ifname: str) -> str:
    """ s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
     return socket.inet_ntoa(fcntl.ioctl(
         s.fileno(),
         0x8915,  # SIOCGIFADDR
         struct.pack('256s', ifname[:15])
     )[20:24])
     """
    return socket.gethostname()


def countnonoverlappingrematches(pattern, thestring):
    return re.subn(pattern, '', thestring)[1]


def getIP() -> str:
    return get_ip_address('eth0')  # '192.168.0.110'


def get_plot_progress(line_count: int) -> float:
    phase1_line_end = PROGRESS_BAR.get('phase1_line_end')
    phase2_line_end = PROGRESS_BAR.get('phase2_line_end')
    phase3_line_end = PROGRESS_BAR.get('phase3_line_end')
    phase4_line_end = PROGRESS_BAR.get('phase4_line_end')
    phase1_weight = PROGRESS_BAR.get('phase1_weight')
    phase2_weight = PROGRESS_BAR.get('phase2_weight')
    phase3_weight = PROGRESS_BAR.get('phase3_weight')
    phase4_weight = PROGRESS_BAR.get('phase4_weight')
    progress = 0
    if line_count > phase1_line_end:
        progress += phase1_weight
    else:
        progress += phase1_weight * (line_count / phase1_line_end)
        return progress
    if line_count > phase2_line_end:
        progress += phase2_weight
    else:
        progress += phase2_weight * ((line_count - phase1_line_end) / (phase2_line_end - phase1_line_end))
        return progress
    if line_count > phase3_line_end:
        progress += phase3_weight
    else:
        progress += phase3_weight * ((line_count - phase2_line_end) / (phase3_line_end - phase2_line_end))
        return progress
    if line_count > phase4_line_end:
        progress += phase4_weight
    else:
        progress += phase4_weight * ((line_count - phase3_line_end) / (phase4_line_end - phase3_line_end))
    return progress



import os
import re
import time
from datetime import datetime

from ..util.plot_util import parse_chia_plot_time, get_plot_progress


class LogFile:
    """
    basic checking and analyzer for this log file
    """

    def __init__(self, logfilename: str):
        self.path = logfilename
        self._total_plots = 0
        self._total_T = 0
        self._pid = 0
        self._produced = 0
        self._start_time = 0
        self._current_plot_id = '--------'
        self._phase = (0, 0)
        self._phase_time = dict()
        self._progress = 0
        self._disk_tight = False

    @property
    def getPhase(self):
        return self._phase

    @property
    def getStartTime(self):
        return self._start_time

    @property
    def getPid(self) -> int:
        return self._pid

    @property
    def getPlotIdFull(self) -> str:
        return self._current_plot_id

    @property
    def getPlotIdShort(self) -> str:
        return self._current_plot_id[:8]

    @property
    def getProductionTPlotSize(self) -> float:
        return self._total_T

    @property
    def getProgress(self) -> float:
        return self._progress

    @property
    def completedJobs(self):
        return self._produced

    # Initialize data that needs to be loaded from the logfile
    def init_logfile(self):
        """
        Read plot ID and job start time from logfile.  Return true if we
           find all the info as expected, false otherwise
        """

        assert self.path
        # Try reading for a while; it can take a while for the job to get started as it scans
        # existing plot dirs (especially if they are NFS).

        found_id = False
        found_log = False
        plots = 0

        for attempt_number in range(1):
            plots = 0
            with open(self.path, 'r') as f:

                for line in f:
                    m = re.match('^ID: ([0-9a-f]*)', line)
                    if m:
                        self._current_plot_id = m.group(1)
                        found_id = True

                    m = re.match(r'^Starting phase 1/4:.*\.\.\. (.*)', line)
                    if m:
                        # Mon Nov  2 08:39:53 2020
                        self._start_time = parse_chia_plot_time(m.group(1))
                        found_log = True
                        # continue and looking for the last

                    m = re.match(r'^Renamed final file from', line)
                    if m:
                        plots += 1

                filedat = f.read()
                lin_count = (filedat.count('\n') + 1)

            if found_id and found_log:
                break  # Stop trying
            else:
                time.sleep(1)  # Sleep and try again

        # If we couldn't find the line in the logfile, the job is probably just getting started
        # (and being slow about it).  In this case, use the last metadata change as the start time.
        # TODO: we never come back to this; e.g. plot_id may remain uninitialized.
        # TODO: should we just use the process start time instead?

        if not found_log:
            self._start_time = datetime.fromtimestamp(os.path.getctime(self.path))
        self._produced = plots
        self._progress = get_plot_progress(lin_count)
        # Load things from logfile that are dynamic
        self.updatePhases()
        # self.updateGeneratePlots()

    def updatePhases(self):
        assert self.path

        # Map from phase number to subphase number reached in that phase.
        # Phase 1 subphases are <started>, table1, table2, ...
        # Phase 2 subphases are <started>, table7, table6, ...
        # Phase 3 subphases are <started>, tables1&2, tables2&3, ...
        # Phase 4 subphases are <started>
        phase_subphases = {}
        plots = 0
        phase = 0
        plotpool = 0
        disk_space = False
        with open(self.path, 'r') as f:

            for line in f:

                m = re.match('^ID: ([0-9a-f]*)', line)
                if m:
                    self._current_plot_id = m.group(1)

                # "Starting phase 1/4: Forward Propagation into tmp files... Sat Oct 31 11:27:04 2020"
                m = re.match(r'^Starting phase (\d).*', line)
                if m:
                    phase = int(m.group(1))
                    phase_subphases[phase] = 0

                # Phase 1: "Computing table 2"
                m = re.match(r'^Computing table (\d).*', line)
                if m:
                    phase_subphases[1] = max(phase_subphases[1], int(m.group(1)))

                # Phase 2: "Backpropagating on table 2"
                m = re.match(r'^Backpropagating on table (\d).*', line)
                if m:
                    phase_subphases[2] = max(phase_subphases[2], 7 - int(m.group(1)))

                # Phase 3: "Compressing tables 4 and 5"
                m = re.match(r'^Compressing tables (\d) and (\d).*', line)
                if m:
                    phase_subphases[3] = max(phase_subphases[3], int(m.group(1)))

                # "Time for phase 1 = 22796.7 seconds. CPU (98%) Tue Sep 29 17:57:19 2020"
                for phase_t in ['1', '2', '3', '4']:
                    m = re.match(r'^Time for phase ' + phase_t + ' = (\d+.\d+) seconds..*', line)
                    if m:
                        self._phase_time.setdefault(f"ph{phase_t}", float(m.group(1)))

                # Total time = 49487.1 seconds. CPU (97.26%) Wed Sep 30 01:22:10 2020
                # m = re.match(r'^Total time = (\d+.\d+) seconds.*', line)
                # if m:
                # data.setdefault(key, {}).setdefault('total time', []).append(float(m.group(1)))
                m = re.match(r'^Final File size: (\d+.\d+).*', line)
                if m:
                    size = float(m.group(1))
                    plotpool = plotpool + size

                m = re.match(r'^Renamed final file from', line)
                if m:
                    phase = 0
                    phase_subphases = {}
                    plots += 1

            lines = f.readlines()
            lastlns = lines[-1:]
            if len(lastlns) > 0:
                m = re.match(r'^Only wrote', lastlns[0])
                if m:
                    disk_space = True

        if phase_subphases:
            phase = max(phase_subphases.keys())
            self._phase = (phase, phase_subphases[phase])
        else:
            self._phase = (0, 0)

        self._produced = plots

        if plotpool > 0:
            self._total_T = plotpool / 1000

        self._disk_tight = disk_space

    @property
    def disk_confirm(self):
        return self._disk_tight

    def print(self):
        print(f"total T size: {self._total_T}")
        print(f"current plot id: {self._current_plot_id}")
        print(f"pid: {self._pid}")
        print(f"produced plots: {self._produced}")



#!/usr/bin/env python3
import re

from .plotinfo import PlotInfo


class PlotLogParser:
    """Parser for a finished plotting job"""

    def parse(self, filename: str) -> PlotInfo:
        """Parses a single log and returns its info"""
        entry = PlotInfo()

        matchers = [
            self.ignore_line,
            self.plot_id,
            self.plot_start_date,
            self.plot_size,
            self.buffer_size,
            self.buckets,
            self.threads,
            self.plot_dirs,
            self.phase1_duration,
            self.phase2_duration,
            self.phase3_duration,
            self.phase4_duration,
            self.total_time,
            self.copy_time,
            self.filename
        ]

        with open(filename, 'r') as f:
            for line in f:
                for matcher in matchers:
                    if matcher(line, entry):
                        break

        return entry

    # ID: 3eb8a37981de1cc76187a36ed947ab4307943cf92967a7e166841186c7899e24
    def plot_id(self, line: str, entry: PlotInfo) -> bool:
        m = re.search(r'^ID: (.+)$', line)
        if m:
            entry.plot_id = m.group(1)
        return m != None

    # Renamed final file from "/farm/wagons/801/abc.plot.2.tmp" to "/farm/wagons/801/abc.plot"
    def filename(self, line: str, entry: PlotInfo) -> bool:
        m = re.search(r'^Renamed final file from ".+" to "(.+)"', line)
        if m:
            entry.filename = m.group(1)
        return m != None

    # Time for phase 1 = 17571.981 seconds. CPU (178.600%) Sun Apr  4 23:53:42 2021
    def phase1_duration(self, line: str, entry: PlotInfo) -> bool:
        m = re.search(r"^Time for phase 1 = (\d+\.\d+) seconds", line)
        if m:
            entry.phase1_duration_raw = float(m.group(1))
        return m != None

    # Time for phase 2 = 6911.621 seconds. CPU (71.780%) Mon Apr  5 01:48:54 2021
    def phase2_duration(self, line: str, entry: PlotInfo) -> bool:
        m = re.search(r"^Time for phase 2 = (\d+\.\d+) seconds", line)
        if m:
            entry.phase2_duration_raw = float(m.group(1))
        return m != None

    # Time for phase 3 = 14537.188 seconds. CPU (82.730%) Mon Apr  5 05:51:11 2021
    def phase3_duration(self, line: str, entry: PlotInfo) -> bool:
        m = re.search(r"^Time for phase 3 = (\d+\.\d+) seconds", line)
        if m:
            entry.phase3_duration_raw = float(m.group(1))
        return m != None

    # Time for phase 4 = 924.288 seconds. CPU (86.810%) Mon Apr  5 06:06:35 2021
    def phase4_duration(self, line: str, entry: PlotInfo) -> bool:
        m = re.search(r"^Time for phase 4 = (\d+\.\d+) seconds", line)
        if m:
            entry.phase4_duration_raw = float(m.group(1))
        return m != None

    # Total time = 39945.080 seconds. CPU (123.100%) Mon Apr  5 06:06:35 2021
    def total_time(self, line: str, entry: PlotInfo) -> bool:
        m = re.search(r"^Total time = (\d+\.\d+) seconds", line)
        if m:
            entry.total_time_raw = float(m.group(1))
        return m != None

    # Copy time = 501.696 seconds. CPU (23.860%) Sun May  9 22:52:41 2021
    def copy_time(self, line: str, entry: PlotInfo) -> bool:
        m = re.search(r"^Copy time = (\d+\.\d+) seconds", line)
        if m:
            entry.copy_time_raw = float(m.group(1))
        return m != None

    # Starting plotting progress into temporary dirs: /farm/yards/901 and /farm/yards/901
    def plot_dirs(self, line: str, entry: PlotInfo) -> bool:
        m = re.search(r"^Starting plotting progress into temporary dirs: (.+) and (.+)$", line)
        if m:
            entry.tmp_dir1 = m.group(1)
            entry.tmp_dir2 = m.group(2)
        return m != None

    # Using 4 threads of stripe size 65536
    def threads(self, line: str, entry: PlotInfo) -> bool:
        m = re.search(r"^Using (\d+) threads of stripe size (\d+)", line)
        if m:
            entry.threads = int(m.group(1))
        return m != None

    # "^Using (\\d+) buckets"
    def buckets(self, line: str, entry: PlotInfo) -> bool:
        m = re.search(r"^Using (\d+) buckets", line)
        if m:
            entry.buckets = int(m.group(1))
        return m != None

    # Buffer size is: 4000MiB
    def buffer_size(self, line: str, entry: PlotInfo) -> bool:
        m = re.search(r"^Buffer size is: (\d+)MiB", line)
        if m:
            entry.buffer = int(m.group(1))
        return m != None

    # Plot size is: 32
    def plot_size(self, line: str, entry: PlotInfo) -> bool:
        m = re.search(r'^Plot size is: (\d+)', line)
        if m:
            entry.plot_size = int(m.group(1))
        return m != None

    # Starting phase 1/4: Forward Propagation into tmp files... Sun May  9 17:36:12 2021
    def plot_start_date(self, line: str, entry: PlotInfo) -> bool:
        m = re.search(r'^Starting phase 1/4: Forward Propagation into tmp files\.\.\. (.+)', line)
        if m:
            entry.started_at = m.group(1)
        return m != None

    # Ignore lines starting with Bucket
    # Bucket 0 uniform sort. Ram: 3.250GiB, u_sort min: 0.563GiB, qs min: 0.281GiB.
    def ignore_line(self, line: str, _: PlotInfo) -> bool:
        return re.search(r'^\tBucket', line)



# TODO do we use all these?
import contextlib
import os
from datetime import datetime

import psutil

from .analyze.pri import LogFile
from .util import plot_util


def job_phases_for_tmpdir(d, all_jobs) -> list:
    """Return phase 2-tuples for jobs running on tmpdir d"""
    return sorted([j.progress() for j in all_jobs if j.tmpdir == d])


def job_phases_for_dstdir(d, all_jobs) -> list:
    """Return phase 2-tuples for jobs outputting to dstdir d"""
    return sorted([j.progress() for j in all_jobs if j.dstdir == d])


def is_plotting_cmdline(cmdline) -> bool:
    return (
            len(cmdline) >= 4
            and 'python' in cmdline[0]
            and cmdline[1].endswith('/chia')
            and 'plots' == cmdline[2]
            and 'create' == cmdline[3]
    )


# This is a cmdline argument fix for https://github.com/ericaltendorf/plotman/issues/41
def cmdline_argfix(cmdline):
    known_keys = 'krbut2dnea'
    for i in cmdline:
        # If the argument starts with dash and a known key and is longer than 2,
        # then an argument is passed with no space between its key and value.
        # This is POSIX compliant but the arg parser was tripping over it.
        # In these cases, splitting that item up in separate key and value
        # elements results in a `cmdline` list that is correctly formatted.
        if i[0] == '-' and i[1] in known_keys and len(i) > 2:
            yield i[0:2]  # key
            yield i[2:]  # value
        else:
            yield i


# TODO: be more principled and explicit about what we cache vs. what we look up
# dynamically from the logfile
class Job:
    """
    Represents a plotter job

    """

    # These are constants, not updated during a run.
    k = 0
    r = 0
    u = 0
    b = 0
    n = 0  # probably not used
    tmpdir = ''
    tmp2dir = ''
    dstdir = ''
    _logpath = ''

    proc = None  # will get a psutil.Process
    help = False

    # These are dynamic, cached, and need to be updated periodically
    phase = (None, None)  # Phase/subphase

    last_updated_time_in_min = 0

    @staticmethod
    def genTasks(logroot, cached_jobs=()) -> list:
        """
        Return a list of running plot jobs.  If a cache of preexisting jobs is provided,
        reuse those previous jobs without updating their information.  Always look for
        new jobs not already in the cache. If the job status is already finish or the storage
        is Zero then we need to make new
        job scanning
        """
        jobs = []
        cached_jobs_by_pid = {j.proc.pid: j for j in cached_jobs}

        for proc in psutil.process_iter(['pid', 'cmdline']):
            # Ignore processes which most likely have terminated between the time of
            # iteration and data access.
            with contextlib.suppress(psutil.NoSuchProcess, psutil.AccessDenied):
                if is_plotting_cmdline(proc.cmdline()):
                    if proc.pid in cached_jobs_by_pid.keys():
                        cached_job = cached_jobs_by_pid[proc.pid]
                        # checking the status for more actions
                        # check for dead locks - Only wrote ...
                        # Only wrote 14708736 of 38859264 bytes at offset 310874112 to "/mnt/nvme/temp/plot-k32-2021-05-25-20-47-f56ddde3daa4cbbf2c7a3a4c79d70d259b26a622ed69ee64df199d253d9e019d.plot.p2.t5.sort_bucket_013.tmp"with length 325582848. Error 1. Retrying in five minutes.
                        if cached_job.get_tmp_usage() == 0:
                            try:
                                job1 = Job(proc, logroot)
                                if not job1.help:
                                    jobs.append(job1)
                            except Exception as e:
                                jobs.append(cached_job)
                        else:
                            jobs.append(cached_job)
                    else:
                        try:
                            job2 = Job(proc, logroot)
                            if not job2.help:
                                jobs.append(job2)
                        except Exception as e:
                            print("Error:", e)
                            pass

        return jobs

    def __init__(self, proc, logroot):
        '''Initialize from an existing psutil.Process object.  must know logroot in order to understand open files'''
        self.proc = proc
        self.zLogFile = None

        with self.proc.oneshot():
            # Parse command line args
            args = self.proc.cmdline()
            assert len(args) > 4
            assert 'python' in args[0]
            assert 'chia' in args[1]
            assert 'plots' == args[2]
            assert 'create' == args[3]
            args_iter = iter(cmdline_argfix(args[4:]))
            for arg in args_iter:
                val = None if arg in {'-e', '--nobitfield', '-h', '--help', '--override-k'} else next(args_iter)
                if arg in {'-k', '--size'}:
                    self.k = val
                elif arg in {'-r', '--num_threads'}:
                    self.r = val
                elif arg in {'-b', '--buffer'}:
                    self.b = val
                elif arg in {'-u', '--buckets'}:
                    self.u = val
                elif arg in {'-t', '--tmp_dir'}:
                    self.tmpdir = val
                elif arg in {'-2', '--tmp2_dir'}:
                    self.tmp2dir = val
                elif arg in {'-d', '--final_dir'}:
                    self.dstdir = val
                elif arg in {'-n', '--num'}:
                    self.n = val
                elif arg in {'-h', '--help'}:
                    self.help = True
                elif arg in {'-e', '--nobitfield', '-f', '--farmer_public_key', '-p', '--pool_public_key', '-a', '--alt_fingerprint'}:
                    pass
                    # TODO: keep track of these
                elif arg == '--override-k':
                    pass
                else:
                    print('Warning: unrecognized args: %s %s' % (arg, val))

            # Find logfile (whatever file is open under the log root).  The
            # file may be open more than once, e.g. for STDOUT and STDERR.
            for f in self.proc.open_files():
                if logroot in f.path:
                    if self._logpath:
                        assert self._logpath == f.path
                    else:
                        self._logpath = f.path
                    break

            if self._logpath:
                self.zLogFile = LogFile(self._logpath)
                self.zLogFile.init_logfile()
                self.check_freeze()

            else:
                print('Found plotting process PID {pid}, but could not find '
                      'logfile in its open files:'.format(pid=self.proc.pid))
                for f in self.proc.open_files():
                    print(f.path)

    def check_freeze(self) -> None:
        assert self._logpath
        updatedAt = os.path.getmtime(self._logpath)
        now = datetime.now().timestamp()
        self.last_updated_time_in_min = int((now - updatedAt) / 60)

    def progress(self):
        """
        Return a 2-tuple with the job phase and subphase (by reading the logfile)
        """
        return self.zLogFile.getPhase

    def isJobStartWithId(self, shortID: str) -> bool:
        if self.zLogFile.getPlotIdFull.startswith(shortID):
            return True
        else:
            return False

    @property
    def getLogPath(self) -> str:
        return self._logpath

    @property
    def plot_id(self) -> str:
        return self.zLogFile.getPlotIdFull

    @property
    def plot_id_prefix(self) -> str:
        return self.zLogFile.getPlotIdShort

    # TODO: make this more useful and complete, and/or make it configurable
    def status_str_long(self) -> str:
        return '{plot_id}\nk={k} r={r} b={b} u={u}\npid:{pid}\ntmp:{tmp}\ntmp2:{tmp2}\ndst:{dst}\nlogfile:{logfile}'.format(
            plot_id=self.plot_id_prefix,
            k=self.k,
            r=self.r,
            b=self.b,
            u=self.u,
            pid=self.proc.pid,
            tmp=self.tmpdir,
            tmp2=self.tmp2dir,
            dst=self.dstdir,
            logfile=self.zLogFile.path
        )

    def get_produced_size(self):
        return self.zLogFile.getProductionTPlotSize

    def get_produced_plots(self):
        return self.zLogFile.completedJobs

    def get_mem_usage(self):
        return self.proc.memory_info().vms  # Total, inc swapped

    def get_tmp_usage(self) -> int:
        total_bytes = 0
        with os.scandir(self.tmpdir) as it:
            for entry in it:
                if self.zLogFile.getPlotIdFull in entry.name:
                    try:
                        total_bytes += entry.stat().st_size
                    except FileNotFoundError:
                        # The file might disappear; this being an estimate we don't care
                        pass
        return total_bytes

    def get_run_status(self) -> str:
        """
        Running, suspended, etc.
        """
        status = self.proc.status()
        if status == psutil.STATUS_RUNNING:
            return 'RUN'
        elif status == psutil.STATUS_SLEEPING:
            return 'SLP'
        elif status == psutil.STATUS_DISK_SLEEP:
            return 'DSK'
        elif status == psutil.STATUS_STOPPED:
            return 'STP'
        else:
            return self.proc.status()

    def get_time_wall(self) -> int:
        create_time = datetime.fromtimestamp(self.proc.create_time())
        return int((datetime.now() - create_time).total_seconds())

    def get_time_user(self) -> int:
        return int(self.proc.cpu_times().user)

    def get_time_sys(self) -> int:
        return int(self.proc.cpu_times().system)

    def get_time_iowait(self) -> int:
        cpu_times = self.proc.cpu_times()
        iowait = getattr(cpu_times, 'iowait', None)
        if iowait is None:
            return 0

        return int(iowait)

    def suspend(self, reason=''):
        self.proc.suspend()
        self.status_note = reason

    def resume(self):
        self.proc.resume()

    def get_temp_files(self):
        # Prevent duplicate file paths by using set.
        temp_files = set([])
        for f in self.proc.open_files():
            if self.tmpdir in f.path or self.tmp2dir in f.path or self.dstdir in f.path:
                temp_files.add(f.path)
        return temp_files

    def toJson(self) -> dict:
        return dict(
            plot_id=self.zLogFile.getPlotIdFull,
            k=self.k,
            r=self.r,
            b=self.b,
            u=self.u,
            pid=self.proc.pid,
            tmp=self.tmpdir,
            tmp2=self.tmp2dir,
            dst=self.dstdir,
            logfile=self.getLogPath,
            freeze=('YES' if self.isFrozen else 'NO'),
            progress=self.zLogFile.getPlotIdFull
        )

    @property
    def isFrozen(self) -> bool:
        return self.last_updated_time_in_min > 60

    @property
    def isWroteErr(self) -> bool:
        return self.zLogFile.disk_confirm

    @property
    def exportSize(self) -> float:
        return self.zLogFile.getProductionTPlotSize

    @property
    def exportProductionPlots(self) -> int:
        return self.zLogFile.completedJobs

    def cancel(self):
        'Cancel an already running job'
        # We typically suspend the job as the first action in killing it, so it
        # doesn't create more tmp files during death.  However, terminate() won't
        # complete if the job is supsended, so we also need to resume it.
        # TODO: check that this is best practice for killing a job.
        self.proc.resume()
        self.proc.terminate()


def report_jdata(jobs, tmp_prefix='', dst_prefix='') -> list:
    jobsr = list()
    for i, j in enumerate(sorted(jobs, key=Job.get_time_wall)):
        with j.proc.oneshot():
            dictionary = {
                'plotid': j.plot_id_prefix,
                'k': j.k,
                'tmp': plot_util.abbr_path(j.tmpdir, tmp_prefix),
                'dst': plot_util.abbr_path(j.dstdir, dst_prefix),
                'wall': plot_util.time_format(j.get_time_wall()),
                'phase': plot_util.phase_str(j.progress()),
                'tmpdisk': plot_util.human_format(j.get_tmp_usage(), 0),
                'pid': j.proc.pid,
                'stat': j.get_run_status(),
                'mem': plot_util.human_format(j.get_mem_usage(), 1),
                'user': plot_util.time_format(j.get_time_user()),
                'sys': plot_util.time_format(j.get_time_sys()),
                'io': plot_util.time_format(j.get_time_iowait()),
                'freezed': plot_util.is_freezed(j),
                'logfile': os.path.basename(j.getLogPath)
            }
            jobsr.append(dictionary)

    return jobsr



import concurrent.futures
import datetime
import glob
import operator
import os
import subprocess

import pendulum
import pkg_resources
import psutil

from .configuration import Scheduling, Directories, Plotting
from .farmplot import FarmPlot
from .job import Job, job_phases_for_tmpdir
from .manager import phases_permit_new_job
from .util import plot_util
# Plotman libraries
from .util.plot_util import getIP, isSpaceCritical

MIN = 60  # Seconds
HR = 3600  # Seconds
MAX_PLOT_SIZE = 332  # Minimum gb required for k32 plot
MAX_AGE = 1000_000_000  # Arbitrary large number of seconds


def terminate(job: Job) -> None:
    job.suspend()
    temp_files = job.get_temp_files()
    job.cancel()
    for f in temp_files:
        os.remove(f)


class MintJ:
    def __init__(self, plotting_cfg: Plotting):
        self.parallel = 0
        self.cpu_clock = 0
        self.youngest_job_age = 0
        self.global_stagger = 0
        self.global_max_jobs = 0
        self.global_stagger_m = 0
        self.wait_reason = None
        self.aging_reason = None
        self.jobs = None
        self.plotdaemon = False
        self.schedule_x = None
        self.dir_cfg_x = None
        self.plotcfg_x = plotting_cfg
        self.bulkn = 0

        self.disk_bytes_read_last = 0
        self.disk_bytes_write_last = 0
        self.net_bytes_read_last = 0
        self.net_bytes_write_last = 0
        self.iowait_last = 0

    def Upcfg(self, schedule: Scheduling, plotting_cfg: Plotting):
        self.schedule_x = schedule
        self.plotcfg_x = plotting_cfg
        self.global_max_jobs = schedule.global_max_jobs
        self.global_stagger_m = schedule.global_stagger_m
        self.youngest_job_age = min(self.jobs, key=Job.get_time_wall).get_time_wall() if self.jobs else MAX_AGE
        self.global_stagger = int(self.global_stagger_m[self.cpu_clock] * MIN)
        self.parallel = schedule.parallel

    def GenJobs(self, dir_cfg: Directories):
        self.dir_cfg_x = dir_cfg
        self.jobs = Job.genTasks(dir_cfg.log)

    def CacheGenJobs(self, dir_cfg: Directories):
        self.jobs = Job.genTasks(dir_cfg.log, cached_jobs=self.jobs)

    @property
    def WaitLog(self) -> str:
        return self.wait_reason

    @property
    def LsJobs(self) -> list:
        return self.jobs

    @property
    def isClockReady(self) -> bool:
        return self.youngest_job_age < self.global_stagger

    @property
    def isCreateNewJobParallelReady(self) -> bool:
        return self.youngest_job_age > self.global_stagger or len(self.jobs) == 0

    @property
    def NewLogFile(self) -> str:
        return os.path.join(self.dir_cfg_x.log, pendulum.now().isoformat(timespec='microseconds').replace(':', '_') + '.log')

    def JobCreate(self):
        if psutil.cpu_percent(interval=10) > 99:
            self.wait_reason = 'cpu optimized!'

        elif self.isClockReady and self.parallel <= 1:
            self.wait_reason = 'stagger (%ds/%ds)' % (self.youngest_job_age, self.global_stagger)

        elif len(self.jobs) >= self.global_max_jobs:
            self.wait_reason = 'max jobs (%d)' % self.global_max_jobs

        else:

            tmp_to_all_phases = [(d, job_phases_for_tmpdir(d, self.jobs)) for d in self.dir_cfg_x.tmp]

            eligible = [(d, phases) for (d, phases) in tmp_to_all_phases
                        if phases_permit_new_job(phases, d, self.schedule_x, self.dir_cfg_x)]

            rankable = [(d, phases[0]) if phases else (d, (999, 999)) for (d, phases) in eligible]

            if not eligible:
                self.wait_reason = 'no eligible tempdirs'
            else:
                # Plot to oldest tmpdir.
                tmpdir = max(rankable, key=operator.itemgetter(1))[0]
                # Select the dst dir least recently selected

                if plot_util.checkTempSpace(int(self.plotcfg_x.k), tmpdir):

                    logfile = self.NewLogFile

                    plot_args = ['chia', 'plots', 'create',
                                 '-k', str(self.plotcfg_x.k),
                                 '-r', str(self.plotcfg_x.n_threads),
                                 '-u', str(self.plotcfg_x.n_buckets),
                                 '-b', str(self.plotcfg_x.job_buffer),
                                 '-n 32',
                                 '-t', tmpdir,
                                 '-d', tmpdir]

                    if self.plotcfg_x.e:
                        plot_args.append('-e')

                    if self.plotcfg_x.farmer_pk is not None:
                        plot_args.append('-f')
                        plot_args.append(self.plotcfg_x.farmer_pk)

                    if self.plotcfg_x.pool_pk is not None:
                        plot_args.append('-p')
                        plot_args.append(self.plotcfg_x.pool_pk)

                    if self.dir_cfg_x.tmp2 is not None:
                        plot_args.append('-2')
                        plot_args.append(self.dir_cfg_x.tmp2)

                    logmsg = ('Starting plot job: %s ; logging to %s' % (' '.join(plot_args), logfile))

                    self.wait_reason = logmsg

                    try:
                        open_log_file = open(logfile, 'x')
                    except FileExistsError:
                        # The desired log file name already exists.  Most likely another
                        # plotman process already launched a new process in response to
                        # the same scenario that triggered us.  Let's at least not
                        # confuse things further by having two plotting processes
                        # logging to the same file.  If we really should launch another
                        # plotting process, we'll get it at the next check cycle anyways.
                        self.wait_reason = (
                            f'Plot log file already exists, skipping attempt to start a'
                            f' new plot: {logfile!r}'
                        )
                        return False

                    # start_new_sessions to make the job independent of this controlling tty.
                    p = subprocess.Popen(plot_args,
                                         stdout=open_log_file,
                                         stderr=subprocess.STDOUT,
                                         start_new_session=True)

                    psutil.Process(p.pid).nice(0)

                    return True
                else:
                    self.wait_reason = f'disk space is not enough! there is only {plot_util.availableSpace(tmpdir)} available in {tmpdir}.'

        return False

    def PlotDaemon(self):
        self.plotdaemon = True

    def SpaceManagement(self) -> None:
        for d in self.dir_cfg_x.tmp:
            ls = []
            if isSpaceCritical(d):
                for j in self.jobs:
                    if j.isFrozen is True:
                        size = plot_util.human_format(j.get_tmp_usage(), 0)
                        print(f'remove from frozen job for space, with temp size {size} [{j.plot_id}]')
                        with concurrent.futures.ThreadPoolExecutor(max_workers=2) as executor:
                            executor.submit(terminate, j)

                    elif j.isWroteErr is True or plot_util.is_phase_start(j.progress()) is True:
                        ls.append(j)

            if len(ls) > 0:
                for i, j in enumerate(sorted(ls, key=Job.get_tmp_usage)):
                    size = plot_util.human_format(j.get_tmp_usage(), 0)
                    print(f'remove temp files for space from sorted temp size {size} [{j.plot_id}]')
                    with concurrent.futures.ThreadPoolExecutor(max_workers=2) as executor:
                        executor.submit(terminate, j)
                    break

    def ParallelWorker(self):
        for i in range(self.parallel):
            sw = self.bulkn % self.parallel
            g = (self.bulkn - sw) / self.parallel

            if g % 2 == 0:
                self.cpu_clock = 0
            else:
                self.cpu_clock = 1

            started = self.JobCreate()
            self.bulkn = self.bulkn + 1
            if not self.plotdaemon:
                if started:
                    self.wait_reason = '<just started job>'
                    self.CacheGenJobs(self.dir_cfg_x)
            else:
                ts = datetime.datetime.now().strftime('%m-%d %H:%M:%S')
                if self.wait_reason:
                    print('> %s, %s' % (ts, self.wait_reason))
                else:
                    print('start plot > %s' % ts)

    def GenStatus(self, jobs: list) -> dict:
        count1 = len(glob.glob1("/mnt/local/tmp", "*.plot"))
        count2 = len(glob.glob1("/mnt/local/temp", "*.plot"))
        count3 = len(glob.glob1("/mnt/nvme/temp", "*.plot"))
        count4 = len(glob.glob1("/mnt/nvme1n0/temp", "*.plot"))

        fcount = count1 + count2 + count3 + count4
        listplmo = FarmPlot.get_running_moving_jobs()
        nfslist = FarmPlot.get_nfs_details()

        disk = psutil.disk_io_counters()
        disk_bytes_read, disk_bytes_write = disk.read_bytes, disk.write_bytes
        net = psutil.net_io_counters()
        net_bytes_read, net_bytes_write = net.bytes_recv, net.bytes_sent
        iowait = psutil.cpu_times().iowait
        disks = [d.mountpoint for d in psutil.disk_partitions()]
        disks_usage = [psutil.disk_usage(d) for d in disks]
        disks_used = sum(d.used for d in disks_usage)
        disks_total = sum(d.total for d in disks_usage)

        d_info = dict(
            jobls=[i.toJson() for i in jobs],
            historyplots=sum([i.exportProductionPlots for i in jobs]),
            sizet=sum([h.exportSize for h in jobs]),
            plotcount=fcount,
            movingcount=len(listplmo),
            movingdetail=listplmo,

            nfsips=nfslist,
            cpucount=psutil.cpu_count(),
            stamp=int(datetime.datetime.now().timestamp()),
            version=pkg_resources.get_distribution('plotmanx').version,

            cpu_percent=psutil.cpu_percent(),
            memory_percent=psutil.virtual_memory().percent,
            cache_percent=round(psutil.virtual_memory().cached / psutil.virtual_memory().total * 100, 1),
            slab_percent=round(psutil.virtual_memory().slab / psutil.virtual_memory().total * 100, 1),
            swap_percent=psutil.swap_memory().percent,
            disk_percent=round((disks_used / disks_total * 100), 2),
            iowait_percent=round((iowait - self.iowait_last) / psutil.cpu_count() * 100, 1),
            net_read_mb_s='{:,}'.format(int((net_bytes_read - self.net_bytes_read_last) / 1024 / 1024)),
            net_write_mb_s='{:,}'.format(int((net_bytes_write - self.net_bytes_write_last) / 1024 / 1024)),
            disk_read_mb_s='{:,}'.format(int((disk_bytes_read - self.disk_bytes_read_last) / 1024 / 1024)),
            disk_write_mb_s='{:,}'.format(int((disk_bytes_write - self.disk_bytes_write_last) / 1024 / 1024)),
            # lsof='{:,}'.format(int(subprocess.check_output('lsof | wc -l', shell=True).decode())),
            net_fds='{:,}'.format(len(psutil.net_connections())),
            identity=getIP(),
        )

        self.net_bytes_read_last, self.net_bytes_write_last = net_bytes_read, net_bytes_write
        self.disk_bytes_read_last, self.disk_bytes_write_last = disk_bytes_read, disk_bytes_write
        self.iowait_last = iowait

        return d_info



import os

# Constants

# Plotman libraries

MIN = 60  # Seconds
HR = 3600  # Seconds
MAX_PLOT_SIZE = 332  # Minimum gb required for k32 plot
MAX_AGE = 1000_000_000  # Arbitrary large number of seconds


def dstdirs_to_furthest_phase(all_jobs):
    '''Return a map from dst dir to a phase tuple for the most progressed job
       that is emitting to that dst dir.'''
    result = {}
    for j in all_jobs:
        if not j.dstdir in result.keys() or result[j.dstdir] < j.progress():
            result[j.dstdir] = j.progress()
    return result


def dstdirs_to_youngest_phase(all_jobs):
    '''Return a map from dst dir to a phase tuple for the least progressed job
       that is emitting to that dst dir.'''
    result = {}
    for j in all_jobs:
        if not j.dstdir in result.keys() or result[j.dstdir] > j.progress():
            result[j.dstdir] = j.progress()
    return result


def phases_permit_new_job(phases, d, sched_cfg, dir_cfg):
    '''Scheduling logic: return True if it's OK to start a new job on a tmp dir
       with existing jobs in the provided phases.'''

    # 当前磁盘剩余空间小于350g不新增
    # if psutil.disk_usage(d).free / 1024 / 1024 / 1024 < 350:
    #    return False

    # Filter unknown-phase jobs
    phases = [ph for ph in phases if ph[0] is not None and ph[1] is not None]

    if len(phases) == 0:
        return True

    milestone = (sched_cfg.tmpdir_stagger_phase_major, sched_cfg.tmpdir_stagger_phase_minor)
    # tmpdir_stagger_phase_limit default is 1, as declared in configuration.py

    if len([p for p in phases if p < milestone]) >= sched_cfg.tmpdir_stagger_phase_limit:
        return False

    # Limit the total number of jobs per tmp dir. Default to the overall max
    # jobs configuration, but restrict to any configured overrides.
    max_plots = sched_cfg.tmpdir_max_jobs
    if dir_cfg.tmp_overrides is not None and d in dir_cfg.tmp_overrides:
        curr_overrides = dir_cfg.tmp_overrides[d]
        if curr_overrides.tmpdir_max_jobs is not None:
            max_plots = curr_overrides.tmpdir_max_jobs
    if len(phases) >= max_plots:
        return False

    return True


def get_size(dir) -> float:
    total_size = 0

    for dirpath, dirnames, filenames in os.walk(dir):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            # skip if it is symbolic link
            if not os.path.islink(fp):
                total_size += (os.path.getsize(fp)) / 1000000000

    return total_size


def anyTmpDirIsNearlyFull(dir_cfg) -> any:
    for dir in dir_cfg.tmp:
        total_size = get_size(dir)
        if total_size > MAX_PLOT_SIZE:
            return (True, total_size, dir)
    return (False, 0, "")


def select_jobs_by_partial_id(jobs, partial_id) -> list:
    selected = []
    for j in jobs:
        if j.isJobStartWithId(partial_id):
            selected.append(j)
    return selected



import sqlite3

from ..configuration import get_db_path


def commaInt(x: str) -> int:
    return int(x.replace(',', ''))


class SQLX:
    def __init__(self):
        self.con = sqlite3.connect(get_db_path())
        self.cur = self.con.cursor()

    def end(self):
        self.con.commit()
        self.con.close()

    def schema_plan(self):
        self.cur.execute('''CREATE TABLE IF NOT EXISTS systemchia (
                           tid text PRIMARY KEY,
                           block text NOT NULL,
                           ip text NOT NULL
                   );''')

        self.cur.execute('''CREATE TABLE IF NOT EXISTS sysio (
                                   tid INTEGER PRIMARY KEY AUTOINCREMENT,
                                   ip text NOT NULL,
                                   plotc INTEGER NOT NULL,
                                   mvplotc INTEGER NOT NULL,
                                   cpu_count INTEGER NOT NULL,
                                   cpu_percent REAL NOT NULL,
                                   cache_percent REAL NOT NULL,
                                   slab_percent REAL NOT NULL,
                                   swap_percent REAL NOT NULL,
                                   disk_percent REAL NOT NULL,
                                   iowait_percent REAL NOT NULL,
                                   memory_percent REAL NOT NULL,
                                   net_read_mb_s INTEGER NOT NULL,
                                   net_write_mb_s INTEGER NOT NULL,
                                   disk_read_mb_s INTEGER NOT NULL,
                                   disk_write_mb_s INTEGER NOT NULL,

                                   net_fds INTEGER NOT NULL,
                                   version text NOT NULL,
                                   stamp INTEGER NOT NULL

                           );''')

        self.cur.execute('''CREATE TABLE IF NOT EXISTS plot (
                           pcid INTEGER PRIMARY KEY AUTOINCREMENT,
                           plotid text NOT NULL,
                           k INTEGER NOT NULL,
                           r INTEGER NOT NULL,
                           b INTEGER NOT NULL,
                           u INTEGER NOT NULL,
                           pid INTEGER NOT NULL,
                           ip text NOT NULL,
                           time text NOT NULL
                   );''')

    def datainput(self, j: dict, ipremo: str, ts: str) -> None:

        if len(j['jobls']) > 0:

            try:
                for h in j['jobls']:

                    plotid = h['plot_id']

                    #                        if len(plotid) > 8:
                    #                            plotid = h['plot_id'][:8]

                    content_find = f"""

                    SELECT COUNT(*) FROM plot WHERE plotid='{plotid}';
                    """

                    (n,) = self.cur.execute(content_find).fetchone()
                    if int(n) == 0:
                        content_insert = f"""
                            INSERT INTO plot (plotid, k, r, b, u, pid, ip, time)
                            VALUES (
                            '{plotid}', {int(h['k'])}, {int(h['r'])}, {int(h['b'])},
                            {int(h['u'])}, {int(h['pid'])}, '{ipremo}', '{ts}'
                            )
                           ;
                        """

                        self.cur.execute(content_insert)

                        print(f"ID - {plotid}")
                    else:
                        content_update = f"""
                        UPDATE plot 
                        SET time='{ts}'
                        WHERE plotid='{plotid}'
                        ;
                        """
                        self.cur.execute(content_update)

            except sqlite3.OperationalError as r:
                print(f"there is a things that doesnt work. {r}")
        else:
            print("body is not empty")

        content_insert = f"""
        INSERT INTO sysio (
            ip,plotc,mvplotc,cpu_count,cpu_percent,cache_percent,slab_percent,
            swap_percent,disk_percent,iowait_percent,memory_percent,
            net_read_mb_s,net_write_mb_s,disk_read_mb_s,disk_write_mb_s,
            net_fds,version,stamp
        ) VALUES(
            '{ipremo}',
            {int(j['plotcount'])},
            {int(j['movingcount'])},
            {int(j['cpucount'])},
            {float(j['cpu_percent'])},
            {float(j['cache_percent'])},
            {float(j['slab_percent'])},
            {float(j['swap_percent'])},
            {float(j['disk_percent'])},
            {float(j['iowait_percent'])},
            {float(j['memory_percent'])},
            {commaInt(j['net_read_mb_s'])},
            {commaInt(j['net_write_mb_s'])},
            {commaInt(j['disk_read_mb_s'])},
            {commaInt(j['disk_write_mb_s'])},
            {commaInt(j['net_fds'])},
            '{j['version']}',
            {int(j['stamp'])}
        )
        """

        self.cur.execute(content_insert)

    def getNodes(self) -> list:
        block = """
        SELECT * FROM sysio GROUP BY ip
        """

        list = self.cur.execute(block).fetchall()

        return list



import pyfakefs

from plotmanx import plot_util
from plotmanx.plot_util import GB


def test_human_format():
    assert (plot_util.human_format(3442000000, 0) == '3G')
    assert (plot_util.human_format(3542000, 2) == '3.54M')
    assert (plot_util.human_format(354, 0) == '354')
    assert (plot_util.human_format(354, 2) == '354.00')


def test_time_format():
    assert (plot_util.time_format(34) == '34s')
    assert (plot_util.time_format(59) == '59s')
    assert (plot_util.time_format(60) == '0:01')
    assert (plot_util.time_format(119) == '0:01')
    assert (plot_util.time_format(120) == '0:02')
    assert (plot_util.time_format(3694) == '1:01')


def test_split_path_prefix():
    assert (plot_util.split_path_prefix([]) == ('', []))
    assert (plot_util.split_path_prefix(['/a/0', '/b/1', '/c/2']) == ('', ['/a/0', '/b/1', '/c/2']))
    assert (plot_util.split_path_prefix(['/a/b/0', '/a/b/1', '/a/b/2']) == ('/a/b', ['0', '1', '2']))


def test_columns():
    assert (plot_util.column_wrap(list(range(8)), 3, filler='--') ==
            [[0, 3, 6],
             [1, 4, 7],
             [2, 5, '--']])
    assert (plot_util.column_wrap(list(range(9)), 3, filler='--') ==
            [[0, 3, 6],
             [1, 4, 7],
             [2, 5, 8]])
    assert (plot_util.column_wrap(list(range(3)), 1, filler='--') ==
            [[0],
             [1],
             [2]])


def test_list_k32_plots(fs: pyfakefs.fake_filesystem.FakeFilesystem):
    fs.create_file('/t/plot-k32-0.plot', st_size=108 * GB)
    fs.create_file('/t/plot-k32-1.plot', st_size=108 * GB)
    fs.create_file('/t/.plot-k32-2.plot', st_size=108 * GB)
    fs.create_file('/t/plot-k32-3.plot.2.tmp', st_size=108 * GB)
    fs.create_file('/t/plot-k32-4.plot', st_size=100 * GB)
    fs.create_file('/t/plot-k32-5.plot', st_size=108 * GB)

    assert (plot_util.list_k32_plots('/t/') ==
            ['/t/plot-k32-0.plot',
             '/t/plot-k32-1.plot',
             '/t/plot-k32-5.plot'])



import os

from .configuration import Directories


def archive_old_log_files(jobs: list, dir: Directories):
    running_logs = [os.path.basename(jb.getLogPath) for jb in jobs]
    archive_files = []
    for root, dirs, files in os.walk(dir.log, topdown=True):
        for f in files:
            if f[-4:] in ['.log', '.txt']:
                if f not in running_logs:
                    archive_files.append(f)






import attr


@attr.s(repr=True, init=False)
class PlotInfo:
    """Represents the results of a finished plot job"""
    started_at: str = ""
    plot_id: str = ""
    buckets: int = 0
    threads: int = 0
    buffer: int = 0
    plot_size: int = 0
    tmp_dir1: str = ""
    tmp_dir2: str = ""
    phase1_duration_raw: float = 0
    phase2_duration_raw: float = 0
    phase3_duration_raw: float = 0
    phase4_duration_raw: float = 0
    total_time_raw: float = 0
    copy_time_raw: float = 0
    filename: str = ""

    def is_empty(self) -> bool:
        "Data is considered empty if total_time is zero"
        return self.total_time == 0

    # Phase 1 duration
    @property
    def phase1_duration(self) -> int:
        return round(self.phase1_duration_raw)

    @property
    def phase1_duration_minutes(self) -> int:
        return self.duration_to_minutes(self.phase1_duration_raw)

    @property
    def phase1_duration_hours(self) -> float:
        return self.duration_to_hours(self.phase1_duration_raw)

    # Phase 2 duration
    @property
    def phase2_duration(self) -> int:
        return round(self.phase2_duration_raw)

    @property
    def phase2_duration_minutes(self) -> int:
        return self.duration_to_minutes(self.phase2_duration_raw)

    @property
    def phase2_duration_hours(self) -> float:
        return self.duration_to_hours(self.phase2_duration_raw)

    # Phase 3 duration
    @property
    def phase3_duration(self) -> int:
        return round(self.phase3_duration_raw)

    @property
    def phase3_duration_minutes(self) -> int:
        return self.duration_to_minutes(self.phase3_duration_raw)

    @property
    def phase3_duration_hours(self) -> float:
        return self.duration_to_hours(self.phase3_duration_raw)

    # Phase 4 duration
    @property
    def phase4_duration(self) -> int:
        return round(self.phase4_duration_raw)

    @property
    def phase4_duration_minutes(self) -> int:
        return self.duration_to_minutes(self.phase4_duration_raw)

    @property
    def phase4_duration_hours(self) -> float:
        return self.duration_to_hours(self.phase4_duration_raw)

    # Total time
    @property
    def total_time(self) -> int:
        return round(self.total_time_raw)

    @property
    def total_time_minutes(self) -> int:
        return self.duration_to_minutes(self.total_time_raw)

    @property
    def total_time_hours(self) -> float:
        return self.duration_to_hours(self.total_time_raw)

    # Copy time
    @property
    def copy_time(self) -> int:
        return round(self.copy_time_raw)

    @property
    def copy_time_minutes(self) -> int:
        return self.duration_to_minutes(self.copy_time_raw)

    @property
    def copy_time_hours(self) -> float:
        return self.duration_to_hours(self.copy_time_raw)

    def duration_to_minutes(self, duration):
        return round(duration / 60)

    def duration_to_hours(self, duration):
        return round(duration / 60 / 60, 2)



# !/usr/bin/env python3

import os

import xlsxwriter

servers = [61, 62, 63, 64, 66, 67, 68, 69, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52]
storage1 = [236, 237, 238, 239, 240, 241, 242, 243]
storage2 = [101, 102, 103, 104, 105, 106, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130]
dst = "/Users/hesdx/Documents/ipfs/deplotplot/workersv3"

c = 0

tyaml = """

user_interface:
        use_stty_size: True
directories:
        log: /home/ipant/.chia/mainnet
        tmp:
                - /mnt/nvme/temp
                - /mnt/nvme0n1/temp

        dst:
                - /mnt/nfs/storage236/chiaFinalData
                - /mnt/nfs/storage237/chiaFinalData
                - /mnt/nfs/storage238/chiaFinalData
                - /mnt/nfs/storage239/chiaFinalData
                
        archive:
                #rsyncd_module: plots
                #rsyncd_path: /plots
                #rsyncd_bwlimit: 80000  # Bandwidth limit in KB/s
                #rsyncd_host: myfarmer
                #rsyncd_user: chia
                #index: 0
scheduling:
        tmpdir_stagger_phase_major: 2
        tmpdir_stagger_phase_minor: 1
        tmpdir_stagger_phase_limit: 6
        tmpdir_max_jobs: 45
        global_max_jobs: 90
        global_stagger_m: [12,18]
        polling_time_s: 30
        parallel: 6

plotting:
        k: 32
        e: True              # Use -e plotting option
        n_threads: 4         # Threads per job
        n_buckets: 128       # Number of buckets to split data into
        job_buffer: 32000     # Per job memory
        # If specified, pass through to the -f and -p options.  See CLI reference.
        farmer_pk: 861467df6768932f1c2f3a7c00a70f1e111b22b55de42c497fb93fd398eb4fd4b213f399e0930d265ffdc2b5655f96f5
        pool_pk: 8e211cb1118b95ac9da2f31fd159a74993bfcc8ab72bff79c880227185ba2b1f372b8e829bd99f5796eb4f68e41f30a9

apis:
        api_polling_throttle_s: 5
        port: 19451
        target: "121.29.84.81"
        """

mjob = """

if [ ! -d "{target}" ]; then
    sudo mkdir -p {target}
    echo "mount drive 192.168.10.{worker_id}"
    sudo mount -t nfs 192.168.10.{worker_id}:/minerdata {target} -o nolock
fi

"""
ransfar = """
nohup ./plmo /mnt/local/tmp /mnt/nfs/{foldername}/chiaFinalData/ 50000000 &
"""

upgrade = """

cd ~/chia-blockchain/
wget https://gitee.com/Yang007GG_com/plotman/attach_files/711891/download/nvme_linux_0.4.80.zip && unzip -o nvme_linux_0.4.80.zip 
rm nvme_linux_0.4.80.zip


"""


def filename(k) -> str:
    if k < len(storage1):
        u = storage1[k]
        return "storage{}".format(u)
    else:
        if k < len(storage2):
            p = storage2[k - len(storage1)]
            return "ipant{}".format(p)
        else:
            p1 = storage2[k - len(storage1) - len(storage2)]
            return "ipant{}".format(p1)


def workerID(k) -> str:
    if k < len(storage1):
        u = storage1[k]
        return u
    else:
        if k < len(storage2):
            p = storage2[k - len(storage1)]
            return p
        else:
            p1 = storage2[k - len(storage1) - len(storage2)]
            return p1


def appendx(d) -> str:
    return "                - /mnt/nfs/{}/chiaFinalData".format(d)


def nameFile(formatc, workerID: int) -> any:
    km = formatc.format(workerID)
    file = os.path.join(dst, km)
    return file


def mountdisk(d) -> str:
    return "/mnt/nfs/{}".format(d)


class ExcelSheetGen:
    def __init__(self, basepath: str):
        # Create a workbook and add a worksheet.
        self.workbook = xlsxwriter.Workbook(os.path.join(basepath, 'chiaNodes.xlsx'))
        self.worksheet = self.workbook.add_worksheet()

        # Start from the first cell. Rows and columns are zero indexed.
        self.row = 1
        self.col = 0
        self.worksheet.write(self.row, self.col, "connect cmd")
        self.worksheet.write(self.row, self.col + 1, "worker IP")
        self.worksheet.write(self.row, self.col + 2, "path NFS")
        self.worksheet.write(self.row, self.col + 3, "path NFS")
        self.worksheet.write(self.row, self.col + 4, "path NFS")
        self.row += 1

    def dataEnter(self, wid, workerId, folder1, folder2, folder3):
        wtxt = "W{}".format(wid)
        wkidtxt = "192.168.10.{}".format(workerId)
        # Some data we want to write to the worksheet.
        self.worksheet.write(self.row, self.col, wtxt)
        self.worksheet.write(self.row, self.col + 1, wkidtxt)
        self.worksheet.write(self.row, self.col + 2, folder1)
        self.worksheet.write(self.row, self.col + 3, folder2)
        self.worksheet.write(self.row, self.col + 4, folder3)
        self.row += 1

    def endExcel(self):
        # Write a total using a formula.
        # self.worksheet.write(row, 0, 'Total')
        # self.worksheet.write(row, 1, '=SUM(B1:B4)')
        self.workbook.close()


book = ExcelSheetGen(dst)

for i in servers:
    k = list()
    id1 = c * 3
    id2 = c * 3 + 1
    id3 = c * 3 + 2
    d1 = filename(id1)
    d2 = filename(id2)
    d3 = filename(id3)

    k.append(appendx(d1))
    k.append(appendx(d2))
    k.append(appendx(d3))

    s = tyaml.format(list="\n".join(k))
    print("W{} -> worker{} -> {} {} {}".format(c + 1, i, d1, d2, d3))

    with open(nameFile("worker{}.yaml", i), 'w') as f:
        f.write(s)
        f.close()

    with open(nameFile("workerMoverInstall{}.sh", i), 'w') as f:
        f.write("#!/bin/bash\n")
        f.write(mjob.format(target=mountdisk(d1), worker_id=workerID(id1)))
        f.write(mjob.format(target=mountdisk(d2), worker_id=workerID(id2)))
        f.write(mjob.format(target=mountdisk(d3), worker_id=workerID(id3)))
        f.write(ransfar.format(foldername=filename(id1)))
        f.write(ransfar.format(foldername=filename(id2)))
        f.write(ransfar.format(foldername=filename(id3)))
        f.write(upgrade)
        f.close()

    book.dataEnter(c + 1, i, d1, d2, d3)
    print("excel new line")

    c += 1
book.endExcel()
